@@ -93,16 +93,18 @@
 
 #define COPYOUT(sin, sa) do {						      \
 	if ((sa) && ((sa)->sa_family == AF_INET || (sa)->sa_family == 255))   \
-		(sin) = ((struct sockaddr_in *)(void *)(sa))->sin_addr;	      \
+		(sin) = ((const struct sockaddr_in *)(const void *)	      \
+		    (sa))->sin_addr;	      \
 	} while (0)
 
 #define COPYOUT6(sin, sa) do {						      \
 	if ((sa) && ((sa)->sa_family == AF_INET6 || (sa)->sa_family == 255))  \
-		(sin) = ((struct sockaddr_in6 *)(void *)(sa))->sin6_addr;     \
+		(sin) = ((const struct sockaddr_in6 *)(const void *)	      \
+		    (sa))->sin6_addr;     \
 	} while (0)
 
 #ifndef CLLADDR
-#  define CLLADDR(s) ((const char *)((s)->sdl_data + (s)->sdl_nlen))
+#  define CLLADDR(s) (const void *)((s)->sdl_data + (s)->sdl_nlen)
 #endif
 
 struct priv {
@@ -142,7 +144,7 @@ if_opensockets_os(struct dhcpcd_ctx *ctx)
 #endif
 
 #define SOCK_FLAGS	(SOCK_CLOEXEC | SOCK_NONBLOCK)
-	ctx->link_fd = xsocket(PF_ROUTE, SOCK_RAW | SOCK_FLAGS, 0);
+	ctx->link_fd = xsocket(PF_ROUTE, SOCK_RAW | SOCK_FLAGS, AF_UNSPEC);
 #undef SOCK_FLAGS
 	return ctx->link_fd == -1 ? -1 : 0;
 }
@@ -171,7 +173,7 @@ if_linkaddr(struct sockaddr_dl *sdl, const struct interface *ifp)
 #endif
 
 static int
-if_getssid1(int s, const char *ifname, uint8_t *ssid)
+if_getssid1(int s, const char *ifname, void *ssid)
 {
 	int retval = -1;
 #if defined(SIOCG80211NWID)
@@ -263,13 +265,15 @@ if_vimaster(const struct dhcpcd_ctx *ctx, const char *ifname)
 }
 
 static void
-get_addrs(int type, char *cp, struct sockaddr **sa)
+get_addrs(int type, const void *data, const struct sockaddr **sa)
 {
+	const char *cp;
 	int i;
 
+	cp = data;
 	for (i = 0; i < RTAX_MAX; i++) {
 		if (type & (1 << i)) {
-			sa[i] = (struct sockaddr *)cp;
+			sa[i] = (const struct sockaddr *)cp;
 			RT_ADVANCE(cp, sa[i]);
 		} else
 			sa[i] = NULL;
@@ -359,8 +363,15 @@ if_findsa(struct dhcpcd_ctx *ctx, const struct sockaddr *sa)
 #ifdef INET
 const char *if_pfname = "Berkley Packet Filter";
 
+void
+if_closeraw(__unused struct interface *ifp, int fd)
+{
+
+	close(fd);
+}
+
 int
-if_openrawsocket(struct interface *ifp, uint16_t protocol)
+if_openraw(struct interface *ifp, uint16_t protocol)
 {
 	struct ipv4_state *state;
 	int fd = -1;
@@ -440,14 +451,14 @@ if_openrawsocket(struct interface *ifp, uint16_t protocol)
 		goto eexit;
 #endif
 
-	/* Install the DHCP filter */
+	/* Install the filter. */
 	memset(&pf, 0, sizeof(pf));
 	if (protocol == ETHERTYPE_ARP) {
 		pf.bf_insns = UNCONST(arp_bpf_filter);
 		pf.bf_len = arp_bpf_filter_len;
 	} else {
-		pf.bf_insns = UNCONST(dhcp_bpf_filter);
-		pf.bf_len = dhcp_bpf_filter_len;
+		pf.bf_insns = UNCONST(bootp_bpf_filter);
+		pf.bf_len = bootp_bpf_filter_len;
 	}
 	if (ioctl(fd, BIOCSETF, &pf) == -1)
 		goto eexit;
@@ -462,12 +473,11 @@ eexit:
 }
 
 ssize_t
-if_sendrawpacket(const struct interface *ifp, uint16_t protocol,
+if_sendraw(__unused const struct interface *ifp, int fd, uint16_t protocol,
     const void *data, size_t len)
 {
 	struct iovec iov[2];
 	struct ether_header hw;
-	int fd;
 
 	memset(&hw, 0, ETHER_HDR_LEN);
 	memset(&hw.ether_dhost, 0xff, ETHER_ADDR_LEN);
@@ -476,25 +486,20 @@ if_sendrawpacket(const struct interface *ifp, uint16_t protocol,
 	iov[0].iov_len = ETHER_HDR_LEN;
 	iov[1].iov_base = UNCONST(data);
 	iov[1].iov_len = len;
-	fd = ipv4_protocol_fd(ifp, protocol);
 	return writev(fd, iov, 2);
 }
 
 /* BPF requires that we read the entire buffer.
  * So we pass the buffer in the API so we can loop on >1 packet. */
 ssize_t
-if_readrawpacket(struct interface *ifp, uint16_t protocol,
-    void *data, size_t len, int *flags)
+if_readraw(struct interface *ifp, int fd, void *data, size_t len, int *flags)
 {
-	int fd;
 	struct bpf_hdr packet;
 	ssize_t bytes;
-	const unsigned char *payload;
+	const char *payload;
 	struct ipv4_state *state;
 
 	state = IPV4_STATE(ifp);
-	fd = ipv4_protocol_fd(ifp, protocol);
-
 	*flags = 0;
 	for (;;) {
 		if (state->buffer_len == 0) {
@@ -521,10 +526,8 @@ if_readrawpacket(struct interface *ifp, uint16_t protocol,
 next:
 		state->buffer_pos += BPF_WORDALIGN(packet.bh_hdrlen +
 		    packet.bh_caplen);
-		if (state->buffer_pos >= state->buffer_len) {
+		if (state->buffer_pos >= state->buffer_len)
 			state->buffer_len = state->buffer_pos = 0;
-			*flags |= RAW_EOF;
-		}
 		if (bytes != -1)
 			return bytes;
 	}
@@ -532,40 +535,36 @@ next:
 
 
 int
-if_address(const struct interface *ifp, const struct in_addr *address,
-    const struct in_addr *netmask, const struct in_addr *broadcast,
-    int action)
+if_address(unsigned char cmd, const struct ipv4_addr *ia)
 {
 	int r;
 	struct in_aliasreq ifra;
 
 	memset(&ifra, 0, sizeof(ifra));
-	strlcpy(ifra.ifra_name, ifp->name, sizeof(ifra.ifra_name));
+	strlcpy(ifra.ifra_name, ia->iface->name, sizeof(ifra.ifra_name));
 
 #define ADDADDR(var, addr) do {						      \
 		(var)->sin_family = AF_INET;				      \
 		(var)->sin_len = sizeof(*(var));			      \
 		(var)->sin_addr = *(addr);				      \
 	} while (/*CONSTCOND*/0)
-	ADDADDR(&ifra.ifra_addr, address);
-	ADDADDR(&ifra.ifra_mask, netmask);
-	if (action >= 0 && broadcast)
-		ADDADDR(&ifra.ifra_broadaddr, broadcast);
+	ADDADDR(&ifra.ifra_addr, &ia->addr);
+	ADDADDR(&ifra.ifra_mask, &ia->mask);
+	if (cmd == RTM_NEWADDR && ia->brd.s_addr != INADDR_ANY)
+		ADDADDR(&ifra.ifra_broadaddr, &ia->brd);
 #undef ADDADDR
 
-	r = ioctl(ifp->ctx->pf_inet_fd,
-	    action < 0 ? SIOCDIFADDR : SIOCAIFADDR, &ifra);
+	r = ioctl(ia->iface->ctx->pf_inet_fd,
+	    cmd == RTM_DELADDR ? SIOCDIFADDR : SIOCAIFADDR, &ifra);
 	return r;
 }
 
 static int
-if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct rt_msghdr *rtm)
+if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, const struct rt_msghdr *rtm)
 {
-	char *cp;
-	struct sockaddr *sa, *rti_info[RTAX_MAX];
+	const struct sockaddr *sa, *rti_info[RTAX_MAX];
 
-	cp = (void *)(rtm + 1);
-	sa = (void *)cp;
+	sa = (const void *)(rtm + 1);
 	if (sa->sa_family != AF_INET)
 		return -1;
 	if (~rtm->rtm_addrs & (RTA_DST | RTA_GATEWAY))
@@ -583,14 +582,14 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct rt_msghdr *rtm)
 		return -1;
 #endif
 
-	get_addrs(rtm->rtm_addrs, cp, rti_info);
+	get_addrs(rtm->rtm_addrs, sa, rti_info);
 	memset(rt, 0, sizeof(*rt));
 	rt->flags = (unsigned int)rtm->rtm_flags;
 	COPYOUT(rt->dest, rti_info[RTAX_DST]);
 	if (rtm->rtm_addrs & RTA_NETMASK)
-		COPYOUT(rt->net, rti_info[RTAX_NETMASK]);
+		COPYOUT(rt->mask, rti_info[RTAX_NETMASK]);
 	else
-		rt->net.s_addr = INADDR_BROADCAST;
+		rt->mask.s_addr = INADDR_BROADCAST;
 	COPYOUT(rt->gate, rti_info[RTAX_GATEWAY]);
 	COPYOUT(rt->src, rti_info[RTAX_IFA]);
 	rt->mtu = (unsigned int)rtm->rtm_rmx.rmx_mtu;
@@ -689,7 +688,7 @@ if_route(unsigned char cmd, const struct rt *rt)
 		if (subnet == -1) /* unikely */
 			rtm.hdr.rtm_addrs &= ~RTA_IFA;
 	}
-	if (rt->net.s_addr == htonl(INADDR_BROADCAST) &&
+	if (rt->mask.s_addr == htonl(INADDR_BROADCAST) &&
 	    rt->gate.s_addr == htonl(INADDR_ANY))
 	{
 #ifdef RTF_CLONING
@@ -705,7 +704,7 @@ if_route(unsigned char cmd, const struct rt *rt)
 		rtm.hdr.rtm_flags |= RTF_HOST;
 #endif
 	} else if (rt->gate.s_addr == htonl(INADDR_LOOPBACK) &&
-	    rt->net.s_addr == htonl(INADDR_BROADCAST))
+	    rt->mask.s_addr == htonl(INADDR_BROADCAST))
 	{
 		rtm.hdr.rtm_flags |= RTF_HOST | RTF_GATEWAY;
 		/* Going via lo0 so remove the interface flags */
@@ -715,7 +714,7 @@ if_route(unsigned char cmd, const struct rt *rt)
 		rtm.hdr.rtm_addrs |= RTA_NETMASK;
 		if (rtm.hdr.rtm_flags & RTF_STATIC)
 			rtm.hdr.rtm_flags |= RTF_GATEWAY;
-		if (rt->net.s_addr == htonl(INADDR_BROADCAST))
+		if (rt->mask.s_addr == htonl(INADDR_BROADCAST))
 			rtm.hdr.rtm_flags |= RTF_HOST;
 	}
 	if ((cmd == RTM_ADD || cmd == RTM_CHANGE) &&
@@ -741,7 +740,7 @@ if_route(unsigned char cmd, const struct rt *rt)
 	}
 
 	if (rtm.hdr.rtm_addrs & RTA_NETMASK)
-		ADDADDR(&rt->net);
+		ADDADDR(&rt->mask);
 
 	if ((cmd == RTM_ADD || cmd == RTM_CHANGE) &&
 	    (rtm.hdr.rtm_addrs & (RTA_IFP | RTA_IFA)))
@@ -772,7 +771,7 @@ if_route(unsigned char cmd, const struct rt *rt)
 }
 
 int
-if_initrt(struct interface *ifp)
+if_initrt(struct dhcpcd_ctx *ctx)
 {
 	struct rt_msghdr *rtm;
 	int mib[6];
@@ -780,7 +779,7 @@ if_initrt(struct interface *ifp)
 	char *buf, *p, *end;
 	struct rt rt;
 
-	ipv4_freerts(ifp->ctx->ipv4_kroutes);
+	ipv4_freerts(ctx->ipv4_kroutes);
 
 	mib[0] = CTL_NET;
 	mib[1] = PF_ROUTE;
@@ -803,17 +802,17 @@ if_initrt(struct interface *ifp)
 	end = buf + needed;
 	for (p = buf; p < end; p += rtm->rtm_msglen) {
 		rtm = (void *)p;
-		if (if_copyrt(ifp->ctx, &rt, rtm) == 0)
-			ipv4_handlert(ifp->ctx, RTM_ADD, &rt, 1);
+		if (if_copyrt(ctx, &rt, rtm) == 0)
+			ipv4_handlert(ctx, RTM_ADD, &rt, 1);
 	}
 	free(buf);
 	return 0;
 }
 
-#ifdef SIOCGIFAFLAG_IN
 int
 if_addrflags(const struct in_addr *addr, const struct interface *ifp)
 {
+#ifdef SIOCGIFAFLAG_IN
 	struct ifreq ifr;
 	struct sockaddr_in *sin;
 
@@ -825,17 +824,12 @@ if_addrflags(const struct in_addr *addr, const struct interface *ifp)
 	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFAFLAG_IN, &ifr) == -1)
 		return -1;
 	return ifr.ifr_addrflags;
-}
 #else
-int
-if_addrflags(__unused const struct in_addr *addr,
-    __unused const struct interface *ifp)
-{
-
-	errno = ENOTSUP;
+	UNUSED(addr);
+	UNUSED(ifp);
 	return 0;
-}
 #endif
+}
 #endif /* INET */
 
 #ifdef INET6
@@ -870,7 +864,7 @@ ifa_scope(struct sockaddr_in6 *sin, unsigned int ifindex)
 #endif
 
 int
-if_address6(const struct ipv6_addr *ia, int action)
+if_address6(unsigned char cmd, const struct ipv6_addr *ia)
 {
 	struct in6_aliasreq ifa;
 	struct in6_addr mask;
@@ -911,18 +905,15 @@ if_address6(const struct ipv6_addr *ia, int action)
 
 	priv = (struct priv *)ia->iface->ctx->priv;
 	return ioctl(priv->pf_inet6_fd,
-	    action < 0 ? SIOCDIFADDR_IN6 : SIOCAIFADDR_IN6, &ifa);
+	    cmd == RTM_DELADDR ? SIOCDIFADDR_IN6 : SIOCAIFADDR_IN6, &ifa);
 }
 
-
 static int
-if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
+if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, const struct rt_msghdr *rtm)
 {
-	char *cp;
-	struct sockaddr *sa, *rti_info[RTAX_MAX];
+	const struct sockaddr *sa, *rti_info[RTAX_MAX];
 
-	cp = (void *)(rtm + 1);
-	sa = (void *)cp;
+	sa = (const void *)(rtm + 1);
 	if (sa->sa_family != AF_INET6)
 		return -1;
 	if (~rtm->rtm_addrs & (RTA_DST | RTA_GATEWAY))
@@ -939,7 +930,7 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
 		return -1;
 #endif
 
-	get_addrs(rtm->rtm_addrs, cp, rti_info);
+	get_addrs(rtm->rtm_addrs, sa, rti_info);
 	memset(rt, 0, sizeof(*rt));
 	rt->flags = (unsigned int)rtm->rtm_flags;
 	COPYOUT6(rt->dest, rti_info[RTAX_DST]);
@@ -952,12 +943,12 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
 		 * a kernel bug or actually used for something.
 		 * Either way it needs to be zeroed out.
 		 */
-		struct sockaddr_in6 *sin6;
+		const struct sockaddr_in6 *sin6;
 		size_t e, i, final = 0, illegal = 0;
 		const unsigned char *p;
 
-		sin6 = (void *)rti_info[RTAX_NETMASK];
-		rt->net = sin6->sin6_addr;
+		sin6 = (const void *)rti_info[RTAX_NETMASK];
+		rt->mask = sin6->sin6_addr;
 		e = sin6->sin6_len - offsetof(struct sockaddr_in6, sin6_addr);
 		if (e > sizeof(struct in6_addr))
 			e = sizeof(struct in6_addr);
@@ -967,7 +958,7 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
 		{
 			if (final && *p) {
 				illegal = 1;
-				rt->net.s6_addr[i++] = 0x00;
+				rt->mask.s6_addr[i++] = 0x00;
 				continue;
 			}
 			switch (*p & 0xff) {
@@ -988,14 +979,14 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
 				break;
 			}
 			if (!illegal)
-				rt->net.s6_addr[i++] &= *p;
+				rt->mask.s6_addr[i++] &= *p;
 			else
-				rt->net.s6_addr[i++] = 0x00;
+				rt->mask.s6_addr[i++] = 0x00;
 		}
-		while (i < sizeof(rt->net.s6_addr))
-			rt->net.s6_addr[i++] = 0x00;
+		while (i < sizeof(rt->mask.s6_addr))
+			rt->mask.s6_addr[i++] = 0x00;
 	} else
-		ipv6_mask(&rt->net, 128);
+		ipv6_mask(&rt->mask, 128);
 	COPYOUT6(rt->gate, rti_info[RTAX_GATEWAY]);
 	DESCOPE(&rt->gate);
 	rt->mtu = (unsigned int)rtm->rtm_rmx.rmx_mtu;
@@ -1104,7 +1095,7 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 	}
 
 	if (rtm.hdr.rtm_addrs & RTA_NETMASK)
-		ADDADDR(&rt->net);
+		ADDADDR(&rt->mask);
 
 	if (rtm.hdr.rtm_addrs & (RTA_IFP | RTA_IFA)) {
 		rtm.hdr.rtm_index = (unsigned short)rt->iface->index;
@@ -1139,7 +1130,7 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 }
 
 int
-if_initrt6(struct interface *ifp)
+if_initrt6(struct dhcpcd_ctx *ctx)
 {
 	struct rt_msghdr *rtm;
 	int mib[6];
@@ -1147,7 +1138,7 @@ if_initrt6(struct interface *ifp)
 	char *buf, *p, *end;
 	struct rt6 rt;
 
-	ipv6_freerts(&ifp->ctx->ipv6->kroutes);
+	ipv6_freerts(&ctx->ipv6->kroutes);
 
 	mib[0] = CTL_NET;
 	mib[1] = PF_ROUTE;
@@ -1170,8 +1161,8 @@ if_initrt6(struct interface *ifp)
 	end = buf + needed;
 	for (p = buf; p < end; p += rtm->rtm_msglen) {
 		rtm = (void *)p;
-		if (if_copyrt6(ifp->ctx, &rt, rtm) == 0)
-			ipv6_handlert(ifp->ctx, RTM_ADD, &rt);
+		if (if_copyrt6(ctx, &rt, rtm) == 0)
+			ipv6_handlert(ctx, RTM_ADD, &rt);
 	}
 	free(buf);
 	return 0;
@@ -1234,215 +1225,252 @@ if_getlifetime6(struct ipv6_addr *ia)
 }
 #endif
 
-int
-if_managelink(struct dhcpcd_ctx *ctx)
+static void
+if_announce(struct dhcpcd_ctx *ctx, const struct if_announcemsghdr *ifan)
+{
+
+	switch(ifan->ifan_what) {
+	case IFAN_ARRIVAL:
+		dhcpcd_handleinterface(ctx, 1, ifan->ifan_name);
+		break;
+	case IFAN_DEPARTURE:
+		dhcpcd_handleinterface(ctx, -1, ifan->ifan_name);
+		break;
+	}
+}
+
+static void
+if_ifinfo(struct dhcpcd_ctx *ctx, const struct if_msghdr *ifm)
 {
-	/* route and ifwatchd like a msg buf size of 2048 */
-	char msg[2048], *p, *e, *cp;
-	ssize_t bytes;
-	struct rt_msghdr *rtm;
-	struct if_announcemsghdr *ifan;
-	struct if_msghdr *ifm;
-	struct ifa_msghdr *ifam;
-	struct sockaddr *sa, *rti_info[RTAX_MAX];
-	int len;
-	struct sockaddr_dl sdl;
 	struct interface *ifp;
-#ifdef INET
-	struct rt rt;
-#endif
-#ifdef INET6
-	struct rt6 rt6;
-	struct in6_addr ia6, net6;
-	struct sockaddr_in6 *sin6;
-#endif
-#if (defined(INET) && defined(IN_IFF_TENTATIVE)) || defined(INET6)
-	int ifa_flags;
-#endif
+	int state;
 
-	if ((bytes = read(ctx->link_fd, msg, sizeof(msg))) == -1)
-		return -1;
-	e = msg + bytes;
-	for (p = msg; p < e; p += rtm->rtm_msglen) {
-		rtm = (void *)p;
-		switch(rtm->rtm_type) {
-#ifdef RTM_IFANNOUNCE
-		case RTM_IFANNOUNCE:
-			ifan = (void *)p;
-			switch(ifan->ifan_what) {
-			case IFAN_ARRIVAL:
-				dhcpcd_handleinterface(ctx, 1,
-				    ifan->ifan_name);
-				break;
-			case IFAN_DEPARTURE:
-				dhcpcd_handleinterface(ctx, -1,
-				    ifan->ifan_name);
-				break;
-			}
-			break;
-#endif
-		case RTM_IFINFO:
-			ifm = (void *)p;
-			ifp = if_findindex(ctx->ifaces, ifm->ifm_index);
-			if (ifp == NULL)
-				break;
-			switch (ifm->ifm_data.ifi_link_state) {
-			case LINK_STATE_DOWN:
-				len = LINK_DOWN;
-				break;
-			case LINK_STATE_UP:
-				len = LINK_UP;
-				break;
-			default:
-				/* handle_carrier will re-load
-				 * the interface flags and check for
-				 * IFF_RUNNING as some drivers that
-				 * don't handle link state also don't
-				 * set IFF_RUNNING when this routing
-				 * message is generated.
-				 * As such, it is a race ...*/
-				len = LINK_UNKNOWN;
-				break;
-			}
-			dhcpcd_handlecarrier(ctx, len,
-			    (unsigned int)ifm->ifm_flags, ifp->name);
-			break;
-		case RTM_ADD:
-		case RTM_CHANGE:
-		case RTM_DELETE:
-			/* Ignore messages generated by us */
-			if (rtm->rtm_pid == getpid()) {
-				ctx->options &= ~DHCPCD_RTM_PPID;
-				continue;
-			}
-			/* Ignore messages sent by the parent after forking */
-			if ((ctx->options &
-			    (DHCPCD_RTM_PPID | DHCPCD_DAEMONISED)) ==
-			    (DHCPCD_RTM_PPID | DHCPCD_DAEMONISED) &&
-			    rtm->rtm_pid == ctx->ppid)
-			{
-				/* If this is the last successful message sent,
-				 * clear the check flag as it's possible another
-				 * process could re-use the same pid and also
-				 * manipulate therouting table. */
-				if (rtm->rtm_seq == ctx->pseq)
-					ctx->options &= ~DHCPCD_RTM_PPID;
-				continue;
-			}
-			cp = (void *)(rtm + 1);
-			sa = (void *)cp;
-			switch (sa->sa_family) {
-#ifdef INET
-			case AF_INET:
-				if (if_copyrt(ctx, &rt, rtm) == 0)
-					ipv4_handlert(ctx, rtm->rtm_type,&rt,0);
-				break;
-#endif
-#ifdef INET6
-			case AF_INET6:
-				if (~rtm->rtm_addrs & (RTA_DST | RTA_GATEWAY))
-					break;
-				/*
-				 * BSD caches host routes in the
-				 * routing table.
-				 * As such, we should be notified of
-				 * reachability by its existance
-				 * with a hardware address
-				 */
-				if (rtm->rtm_flags & (RTF_HOST)) {
-					get_addrs(rtm->rtm_addrs, cp, rti_info);
-					COPYOUT6(ia6, rti_info[RTAX_DST]);
-					DESCOPE(&ia6);
-					if (rti_info[RTAX_GATEWAY]->sa_family
-					    == AF_LINK)
-						memcpy(&sdl,
-						    rti_info[RTAX_GATEWAY],
-						    sizeof(sdl));
-					else
-						sdl.sdl_alen = 0;
-					ipv6nd_neighbour(ctx, &ia6,
-					    rtm->rtm_type != RTM_DELETE &&
-					    sdl.sdl_alen ?
-					    IPV6ND_REACHABLE : 0);
-					break;
-				}
+	if ((ifp = if_findindex(ctx->ifaces, ifm->ifm_index)) == NULL)
+		return;
+	switch (ifm->ifm_data.ifi_link_state) {
+	case LINK_STATE_DOWN:
+		state = LINK_DOWN;
+		break;
+	case LINK_STATE_UP:
+		state = LINK_UP;
+		break;
+	default:
+		/* handle_carrier will re-load the interface flags and check for
+		 * IFF_RUNNING as some drivers that don't handle link state also
+		 * don't set IFF_RUNNING when this routing message is generated.
+		 * As such, it is a race ...*/
+		state = LINK_UNKNOWN;
+		break;
+	}
+	dhcpcd_handlecarrier(ctx, state,
+	    (unsigned int)ifm->ifm_flags, ifp->name);
+}
 
-				if (if_copyrt6(ctx, &rt6, rtm) == 0)
-					ipv6_handlert(ctx, rtm->rtm_type, &rt6);
-				break;
-#endif
-			}
-			break;
-#ifdef RTM_CHGADDR
-		case RTM_CHGADDR:	/* FALLTHROUGH */
-#endif
-		case RTM_DELADDR:	/* FALLTHROUGH */
-		case RTM_NEWADDR:
-			/* XXX We have no way of knowing who generated these
-			 * messages wich truely sucks because we want to
-			 * avoid listening to our own delete messages. */
-			ifam = (void *)p;
-			ifp = if_findindex(ctx->ifaces, ifam->ifam_index);
-			if (ifp == NULL)
-				break;
-			cp = (void *)(ifam + 1);
-			get_addrs(ifam->ifam_addrs, cp, rti_info);
-			if (rti_info[RTAX_IFA] == NULL)
-				break;
-			switch (rti_info[RTAX_IFA]->sa_family) {
-			case AF_LINK:
-#ifdef RTM_CHGADDR
-				if (rtm->rtm_type != RTM_CHGADDR)
-					break;
-#else
-				if (rtm->rtm_type != RTM_NEWADDR)
-					break;
-#endif
-				memcpy(&sdl, rti_info[RTAX_IFA],
-				    rti_info[RTAX_IFA]->sa_len);
-				dhcpcd_handlehwaddr(ctx, ifp->name,
-				    (const unsigned char*)CLLADDR(&sdl),
-				    sdl.sdl_alen);
-				break;
+static void
+if_rtm(struct dhcpcd_ctx *ctx, const struct rt_msghdr *rtm)
+{
+	const struct sockaddr *sa;
+
+	/* Ignore messages generated by us */
+	if (rtm->rtm_pid == getpid()) {
+		ctx->options &= ~DHCPCD_RTM_PPID;
+		return;
+	}
+
+	/* Ignore messages sent by the parent after forking */
+	if ((ctx->options &
+	    (DHCPCD_RTM_PPID | DHCPCD_DAEMONISED)) ==
+	    (DHCPCD_RTM_PPID | DHCPCD_DAEMONISED) &&
+	    rtm->rtm_pid == ctx->ppid)
+	{
+		/* If this is the last successful message sent,
+		 * clear the check flag as it's possible another
+		 * process could re-use the same pid and also
+		 * manipulate therouting table. */
+		if (rtm->rtm_seq == ctx->pseq)
+			ctx->options &= ~DHCPCD_RTM_PPID;
+		return;
+	}
+
+	sa = (const void *)(rtm + 1);
+	switch (sa->sa_family) {
 #ifdef INET
-			case AF_INET:
-			case 255: /* FIXME: Why 255? */
-				COPYOUT(rt.dest, rti_info[RTAX_IFA]);
-				COPYOUT(rt.net, rti_info[RTAX_NETMASK]);
-				COPYOUT(rt.gate, rti_info[RTAX_BRD]);
-				if (rtm->rtm_type == RTM_NEWADDR) {
-					ifa_flags = if_addrflags(&rt.dest, ifp);
-					if (ifa_flags == -1)
-						break;
-				} else
-					ifa_flags = 0;
-				ipv4_handleifa(ctx, rtm->rtm_type,
-				    NULL, ifp->name,
-				    &rt.dest, &rt.net, &rt.gate, ifa_flags);
-				break;
+	case AF_INET:
+	{
+		struct rt rt;
+
+		if (if_copyrt(ctx, &rt, rtm) == 0)
+			ipv4_handlert(ctx, rtm->rtm_type, &rt, 0);
+		break;
+	}
 #endif
 #ifdef INET6
-			case AF_INET6:
-				sin6 = (void *)rti_info[RTAX_IFA];
-				ia6 = sin6->sin6_addr;
-				DESCOPE(&ia6);
-				sin6 = (void *)rti_info[RTAX_NETMASK];
-				net6 = sin6->sin6_addr;
-				DESCOPE(&net6);
-				if (rtm->rtm_type == RTM_NEWADDR) {
-					ifa_flags = if_addrflags6(&ia6, ifp);
-					if (ifa_flags == -1)
-						break;
-				} else
-					ifa_flags = 0;
-				ipv6_handleifa(ctx, rtm->rtm_type, NULL,
-				    ifp->name, &ia6, ipv6_prefixlen(&net6),
-				    ifa_flags);
-				break;
-#endif
-			}
+	case AF_INET6:
+	{
+		struct rt6 rt6;
+
+		if (~rtm->rtm_addrs & (RTA_DST | RTA_GATEWAY))
+			break;
+		/*
+		 * BSD announces host routes.
+		 * As such, we should be notified of reachability by its
+		 * existance with a hardware address.
+		 */
+		if (rtm->rtm_flags & (RTF_HOST)) {
+			const struct sockaddr *rti_info[RTAX_MAX];
+			struct in6_addr dst6;
+			struct sockaddr_dl sdl;
+
+			get_addrs(rtm->rtm_addrs, rtm + 1, rti_info);
+			COPYOUT6(dst6, rti_info[RTAX_DST]);
+			DESCOPE(&dst6);
+			if (rti_info[RTAX_GATEWAY]->sa_family == AF_LINK)
+				memcpy(&sdl, rti_info[RTAX_GATEWAY],
+				    sizeof(sdl));
+			else
+				sdl.sdl_alen = 0;
+			ipv6nd_neighbour(ctx, &dst6,
+			    rtm->rtm_type != RTM_DELETE && sdl.sdl_alen ?
+			    IPV6ND_REACHABLE : 0);
 			break;
 		}
+
+		if (if_copyrt6(ctx, &rt6, rtm) == 0)
+			ipv6_handlert(ctx, rtm->rtm_type, &rt6);
+		break;
+	}
+#endif
+	}
+}
+
+static void
+if_ifa(struct dhcpcd_ctx *ctx, const struct ifa_msghdr *ifam)
+{
+	struct interface *ifp;
+	const struct sockaddr *rti_info[RTAX_MAX];
+
+	/* XXX We have no way of knowing who generated these
+	 * messages wich truely sucks because we want to
+	 * avoid listening to our own delete messages. */
+	if ((ifp = if_findindex(ctx->ifaces, ifam->ifam_index)) == NULL)
+		return;
+	get_addrs(ifam->ifam_addrs, ifam + 1, rti_info);
+	if (rti_info[RTAX_IFA] == NULL)
+		return;
+	switch (rti_info[RTAX_IFA]->sa_family) {
+	case AF_LINK:
+	{
+		struct sockaddr_dl sdl;
+
+#ifdef RTM_CHGADDR
+		if (ifam->ifam_type != RTM_CHGADDR)
+			break;
+#else
+		if (ifam->ifam_type != RTM_NEWADDR)
+			break;
+#endif
+		memcpy(&sdl, rti_info[RTAX_IFA], rti_info[RTAX_IFA]->sa_len);
+		dhcpcd_handlehwaddr(ctx, ifp->name, CLLADDR(&sdl),sdl.sdl_alen);
+		break;
+	}
+#ifdef INET
+	case AF_INET:
+	case 255: /* FIXME: Why 255? */
+	{
+		const struct sockaddr_in *sin;
+		struct in_addr addr, mask, bcast;
+		int flags;
+
+		sin = (const void *)rti_info[RTAX_IFA];
+		addr.s_addr = sin != NULL && sin->sin_family == AF_INET ?
+		    sin->sin_addr.s_addr : INADDR_ANY;
+		sin = (const void *)rti_info[RTAX_NETMASK];
+		mask.s_addr = sin != NULL && sin->sin_family == AF_INET ?
+		    sin->sin_addr.s_addr : INADDR_ANY;
+		sin = (const void *)rti_info[RTAX_BRD];
+		bcast.s_addr = sin != NULL && sin->sin_family == AF_INET ?
+		    sin->sin_addr.s_addr : INADDR_ANY;
+		if (ifam->ifam_type == RTM_NEWADDR) {
+			if ((flags = if_addrflags(&addr, ifp)) == -1)
+				break;
+		} else
+			flags = 0;
+		ipv4_handleifa(ctx, ifam->ifam_type, NULL, ifp->name,
+		    &addr, &mask, &bcast, flags);
+		break;
+	}
+#endif
+#ifdef INET6
+	case AF_INET6:
+	{
+		struct in6_addr addr6, mask6;
+		const struct sockaddr_in6 *sin6;
+		int flags;
+
+		sin6 = (const void *)rti_info[RTAX_IFA];
+		addr6 = sin6->sin6_addr;
+		DESCOPE(&addr6);
+		sin6 = (const void *)rti_info[RTAX_NETMASK];
+		mask6 = sin6->sin6_addr;
+		DESCOPE(&mask6);
+		if (ifam->ifam_type == RTM_NEWADDR) {
+			if ((flags = if_addrflags6(&addr6, ifp)) == -1)
+				break;
+		} else
+			flags = 0;
+		ipv6_handleifa(ctx, ifam->ifam_type, NULL,
+		    ifp->name, &addr6, ipv6_prefixlen(&mask6), flags);
+		break;
+	}
+#endif
+	}
+}
+
+static void
+if_dispatch(struct dhcpcd_ctx *ctx, const struct rt_msghdr *rtm)
+{
+
+	switch(rtm->rtm_type) {
+#ifdef RTM_IFANNOUNCE
+	case RTM_IFANNOUNCE:
+		if_announce(ctx, (const void *)rtm);
+		break;
+#endif
+	case RTM_IFINFO:
+		if_ifinfo(ctx, (const void *)rtm);
+		break;
+	case RTM_ADD:		/* FALLTHROUGH */
+	case RTM_CHANGE:	/* FALLTHROUGH */
+	case RTM_DELETE:
+		if_rtm(ctx, (const void *)rtm);
+		break;
+#ifdef RTM_CHGADDR
+	case RTM_CHGADDR:	/* FALLTHROUGH */
+#endif
+	case RTM_DELADDR:	/* FALLTHROUGH */
+	case RTM_NEWADDR:
+		if_ifa(ctx, (const void *)rtm);
+		break;
+	}
+}
+
+int
+if_handlelink(struct dhcpcd_ctx *ctx)
+{
+	/* route and ifwatchd like a msg buf size of 2048 */
+	char buf[2048];
+	const char *p, *e;
+	size_t msglen;
+	ssize_t bytes;
+	const struct rt_msghdr *rtm;
+
+	if ((bytes = read(ctx->link_fd, buf, sizeof(buf))) == -1)
+		return -1;
+	e = buf + bytes;
+	for (p = buf; p < e; p += msglen) {
+		rtm = (const void *)p;
+		msglen = rtm->rtm_msglen;
+		if_dispatch(ctx, rtm);
 	}
 	return 0;
 }
