@@ -392,7 +392,7 @@ get_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp, int fd, int flags,
 			r = err_netlink(nlm);
 			if (r == -1)
 				goto eexit;
-			
+
 			if (r == 0 && callback) {
 				r = callback(ctx, ifp, nlm);
 				if (r != 0)
@@ -475,7 +475,7 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
 		rta = RTA_NEXT(rta, len);
 	}
 
-	inet_cidrtoaddr(rtm->rtm_dst_len, &rt->net);
+	inet_cidrtoaddr(rtm->rtm_dst_len, &rt->mask);
 	if (rt->iface == NULL && rt->src.s_addr != INADDR_ANY) {
 		struct ipv4_addr *ap;
 
@@ -517,7 +517,7 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct nlmsghdr *nlm)
 		rt->flags = RTF_REJECT;
 	if (rtm->rtm_scope == RT_SCOPE_HOST)
 		rt->flags |= RTF_HOST;
-	ipv6_mask(&rt->net, rtm->rtm_dst_len);
+	ipv6_mask(&rt->mask, rtm->rtm_dst_len);
 
 	rta = (struct rtattr *)RTM_RTA(rtm);
 	len = RTM_PAYLOAD(nlm);
@@ -885,7 +885,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 }
 
 int
-if_managelink(struct dhcpcd_ctx *ctx)
+if_handlelink(struct dhcpcd_ctx *ctx)
 {
 
 	return get_netlink(ctx, NULL,
@@ -1019,7 +1019,9 @@ nla_next(struct nlattr *nla, size_t *rem)
 	(nla)->nla_len <= rem)
 #define NLA_DATA(nla) ((char *)(nla) + NLA_HDRLEN)
 #define NLA_FOR_EACH_ATTR(pos, head, len, rem) \
-	for (pos = head, rem = len; NLA_OK(pos, rem); pos = nla_next(pos, &(rem)))
+	for (pos = head, rem = len; \
+	     NLA_OK(pos, rem); \
+	     pos = nla_next(pos, &(rem)))
 
 struct nlmg
 {
@@ -1216,8 +1218,15 @@ struct nlmr
 #ifdef INET
 const char *if_pfname = "Packet Socket";
 
+void
+if_closeraw(__unused struct interface *ifp, int fd)
+{
+
+	close(fd);
+}
+
 int
-if_openrawsocket(struct interface *ifp, uint16_t protocol)
+if_openraw(struct interface *ifp, uint16_t protocol)
 {
 	int s;
 	union sockunion {
@@ -1235,14 +1244,14 @@ if_openrawsocket(struct interface *ifp, uint16_t protocol)
 		return -1;
 #undef SF
 
-	/* Install the DHCP filter */
+	/* Install the filter. */
 	memset(&pf, 0, sizeof(pf));
 	if (protocol == ETHERTYPE_ARP) {
 		pf.filter = UNCONST(arp_bpf_filter);
 		pf.len = arp_bpf_filter_len;
 	} else {
-		pf.filter = UNCONST(dhcp_bpf_filter);
-		pf.len = dhcp_bpf_filter_len;
+		pf.filter = UNCONST(bootp_bpf_filter);
+		pf.len = bootp_bpf_filter_len;
 	}
 	if (setsockopt(s, SOL_SOCKET, SO_ATTACH_FILTER, &pf, sizeof(pf)) != 0)
 		goto eexit;
@@ -1268,7 +1277,7 @@ eexit:
 }
 
 ssize_t
-if_sendrawpacket(const struct interface *ifp, uint16_t protocol,
+if_sendraw(const struct interface *ifp, int fd, uint16_t protocol,
     const void *data, size_t len)
 {
 	union sockunion {
@@ -1276,7 +1285,6 @@ if_sendrawpacket(const struct interface *ifp, uint16_t protocol,
 		struct sockaddr_ll sll;
 		struct sockaddr_storage ss;
 	} su;
-	int fd;
 
 	memset(&su, 0, sizeof(su));
 	su.sll.sll_family = AF_PACKET;
@@ -1294,13 +1302,12 @@ if_sendrawpacket(const struct interface *ifp, uint16_t protocol,
 		    &ipv4_bcast_addr, sizeof(ipv4_bcast_addr));
 	} else
 		memset(&su.sll.sll_addr, 0xff, ifp->hwlen);
-	fd = ipv4_protocol_fd(ifp, protocol);
 
 	return sendto(fd, data, len, 0, &su.sa, sizeof(su.sll));
 }
 
 ssize_t
-if_readrawpacket(struct interface *ifp, uint16_t protocol,
+if_readraw(__unused struct interface *ifp, int fd,
     void *data, size_t len, int *flags)
 {
 	struct iovec iov = {
@@ -1318,18 +1325,16 @@ if_readrawpacket(struct interface *ifp, uint16_t protocol,
 #endif
 
 	ssize_t bytes;
-	int fd = -1;
 
 #ifdef PACKET_AUXDATA
 	msg.msg_control = cmsgbuf;
 	msg.msg_controllen = sizeof(cmsgbuf);
 #endif
 
-	fd = ipv4_protocol_fd(ifp, protocol);
 	bytes = recvmsg(fd, &msg, 0);
 	if (bytes == -1)
 		return -1;
-	*flags = RAW_EOF; /* We only ever read one packet */
+	*flags = 0;
 	if (bytes) {
 #ifdef PACKET_AUXDATA
 		for (cmsg = CMSG_FIRSTHDR(&msg);
@@ -1349,9 +1354,7 @@ if_readrawpacket(struct interface *ifp, uint16_t protocol,
 }
 
 int
-if_address(const struct interface *iface,
-    const struct in_addr *address, const struct in_addr *netmask,
-    const struct in_addr *broadcast, int action)
+if_address(unsigned char cmd, const struct ipv4_addr *addr)
 {
 	struct nlma nlm;
 	int retval = 0;
@@ -1359,24 +1362,24 @@ if_address(const struct interface *iface,
 	memset(&nlm, 0, sizeof(nlm));
 	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
 	nlm.hdr.nlmsg_flags = NLM_F_REQUEST;
-	if (action >= 0) {
+	nlm.hdr.nlmsg_type = cmd;
+	if (cmd == RTM_NEWADDR)
 		nlm.hdr.nlmsg_flags |= NLM_F_CREATE | NLM_F_REPLACE;
-		nlm.hdr.nlmsg_type = RTM_NEWADDR;
-	} else
-		nlm.hdr.nlmsg_type = RTM_DELADDR;
-	nlm.ifa.ifa_index = iface->index;
+	nlm.ifa.ifa_index = addr->iface->index;
 	nlm.ifa.ifa_family = AF_INET;
-	nlm.ifa.ifa_prefixlen = inet_ntocidr(*netmask);
+	nlm.ifa.ifa_prefixlen = inet_ntocidr(addr->mask);
 	/* This creates the aliased interface */
 	add_attr_l(&nlm.hdr, sizeof(nlm), IFA_LABEL,
-	    iface->alias, (unsigned short)(strlen(iface->alias) + 1));
+	    addr->iface->alias,
+	    (unsigned short)(strlen(addr->iface->alias) + 1));
 	add_attr_l(&nlm.hdr, sizeof(nlm), IFA_LOCAL,
-	    &address->s_addr, sizeof(address->s_addr));
-	if (action >= 0 && broadcast)
+	    &addr->addr.s_addr, sizeof(addr->addr.s_addr));
+	if (cmd == RTM_NEWADDR)
 		add_attr_l(&nlm.hdr, sizeof(nlm), IFA_BROADCAST,
-		    &broadcast->s_addr, sizeof(broadcast->s_addr));
+		    &addr->brd.s_addr, sizeof(addr->brd.s_addr));
 
-	if (send_netlink(iface->ctx, NULL, NETLINK_ROUTE, &nlm.hdr, NULL) == -1)
+	if (send_netlink(addr->iface->ctx, NULL,
+	    NETLINK_ROUTE, &nlm.hdr, NULL) == -1)
 		retval = -1;
 	return retval;
 }
@@ -1420,14 +1423,14 @@ if_route(unsigned char cmd, const struct rt *rt)
 			nlm.rt.rtm_scope = RT_SCOPE_HOST;
 		else if (rt->gate.s_addr == INADDR_ANY ||
 		    (rt->gate.s_addr == rt->dest.s_addr &&
-			rt->net.s_addr == INADDR_BROADCAST))
+			rt->mask.s_addr == INADDR_BROADCAST))
 			nlm.rt.rtm_scope = RT_SCOPE_LINK;
 		else
 			nlm.rt.rtm_scope = RT_SCOPE_UNIVERSE;
 		nlm.rt.rtm_type = RTN_UNICAST;
 	}
 
-	nlm.rt.rtm_dst_len = inet_ntocidr(rt->net);
+	nlm.rt.rtm_dst_len = inet_ntocidr(rt->mask);
 	add_attr_l(&nlm.hdr, sizeof(nlm), RTA_DST,
 	    &rt->dest.s_addr, sizeof(rt->dest.s_addr));
 	if (cmd == RTM_ADD || cmd == RTM_CHANGE) {
@@ -1474,11 +1477,11 @@ _if_initrt(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 }
 
 int
-if_initrt(struct interface *ifp)
+if_initrt(struct dhcpcd_ctx *ctx)
 {
 	struct nlmr nlm;
 
-	ipv4_freerts(ifp->ctx->ipv4_kroutes);
+	ipv4_freerts(ctx->ipv4_kroutes);
 
 	memset(&nlm, 0, sizeof(nlm));
 	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
@@ -1487,10 +1490,8 @@ if_initrt(struct interface *ifp)
 	nlm.hdr.nlmsg_flags |= NLM_F_REQUEST;
 	nlm.rt.rtm_family = AF_INET;
 	nlm.rt.rtm_table = RT_TABLE_MAIN;
-	add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, ifp->index);
 
-	return send_netlink(ifp->ctx, ifp,
-	    NETLINK_ROUTE, &nlm.hdr, &_if_initrt);
+	return send_netlink(ctx, NULL, NETLINK_ROUTE, &nlm.hdr, &_if_initrt);
 }
 
 int
@@ -1505,7 +1506,7 @@ if_addrflags(__unused const struct in_addr *addr,
 
 #ifdef INET6
 int
-if_address6(const struct ipv6_addr *ia, int action)
+if_address6(unsigned char cmd, const struct ipv6_addr *ia)
 {
 	struct nlma nlm;
 	struct ifa_cacheinfo cinfo;
@@ -1518,11 +1519,9 @@ if_address6(const struct ipv6_addr *ia, int action)
 	memset(&nlm, 0, sizeof(nlm));
 	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
 	nlm.hdr.nlmsg_flags = NLM_F_REQUEST;
-	if (action >= 0) {
+	nlm.hdr.nlmsg_type = cmd;
+	if (cmd == RTM_NEWADDR)
 		nlm.hdr.nlmsg_flags |= NLM_F_CREATE | NLM_F_REPLACE;
-		nlm.hdr.nlmsg_type = RTM_NEWADDR;
-	} else
-		nlm.hdr.nlmsg_type = RTM_DELADDR;
 	nlm.ifa.ifa_index = ia->iface->index;
 	nlm.ifa.ifa_family = AF_INET6;
 #ifdef IPV6_MANAGETEMPADDR
@@ -1547,7 +1546,7 @@ if_address6(const struct ipv6_addr *ia, int action)
 	add_attr_l(&nlm.hdr, sizeof(nlm), IFA_LOCAL,
 	    &ia->addr.s6_addr, sizeof(ia->addr.s6_addr));
 
-	if (action >= 0) {
+	if (cmd == RTM_NEWADDR) {
 		memset(&cinfo, 0, sizeof(cinfo));
 		cinfo.ifa_prefered = ia->prefix_pltime;
 		cinfo.ifa_valid = ia->prefix_vltime;
@@ -1609,7 +1608,7 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 			nlm.rt.rtm_type = RTN_UNICAST;
 	}
 
-	nlm.rt.rtm_dst_len = ipv6_prefixlen(&rt->net);
+	nlm.rt.rtm_dst_len = ipv6_prefixlen(&rt->mask);
 	add_attr_l(&nlm.hdr, sizeof(nlm), RTA_DST,
 	    &rt->dest.s6_addr, sizeof(rt->dest.s6_addr));
 
@@ -1648,11 +1647,11 @@ _if_initrt6(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 }
 
 int
-if_initrt6(struct interface *ifp)
+if_initrt6(struct dhcpcd_ctx *ctx)
 {
 	struct nlmr nlm;
 
-	ipv6_freerts(&ifp->ctx->ipv6->kroutes);
+	ipv6_freerts(&ctx->ipv6->kroutes);
 
 	memset(&nlm, 0, sizeof(nlm));
 	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
@@ -1661,10 +1660,8 @@ if_initrt6(struct interface *ifp)
 	nlm.hdr.nlmsg_flags |= NLM_F_REQUEST;
 	nlm.rt.rtm_family = AF_INET6;
 	nlm.rt.rtm_table = RT_TABLE_MAIN;
-	add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, ifp->index);
 
-	return send_netlink(ifp->ctx, ifp,
-	    NETLINK_ROUTE, &nlm.hdr, &_if_initrt6);
+	return send_netlink(ctx, NULL, NETLINK_ROUTE, &nlm.hdr, &_if_initrt6);
 }
 
 int
