@@ -211,14 +211,13 @@ static void if_learnaddrs(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 			addr = (void *)ifa->ifa_addr;
 			net = (void *)ifa->ifa_netmask;
 			if (ifa->ifa_flags & IFF_POINTOPOINT)
-				brd = (const struct sockaddr_in *)
-				    (void *)ifa->ifa_dstaddr;
+				brd = (void *)ifa->ifa_dstaddr;
 			else
 				brd = (void *)ifa->ifa_broadaddr;
 			ifa_flags = if_addrflags(&addr->sin_addr, ifp);
 			ipv4_handleifa(ctx, RTM_NEWADDR, ifs, ifa->ifa_name,
-				&addr->sin_addr, &net->sin_addr, &brd->sin_addr,
-				ifa_flags);
+				&addr->sin_addr, &net->sin_addr,
+				brd ? &brd->sin_addr : NULL, ifa_flags);
 			break;
 #endif
 #ifdef INET6
@@ -271,16 +270,11 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 	const struct sockaddr_ll *sll;
 #endif
 
-#ifdef GETIFADDRS_AFLINK
 	if (getifaddrs(&ifaddrs) == -1)
 		return NULL;
-#else
-	if (if_getifaddrs(&ifaddrs) == -1)
-		return NULL;
-#endif
-	ifs = malloc(sizeof(*ifs));
-	if (ifs == NULL)
-		return NULL;
+
+	if ((ifs = malloc(sizeof(*ifs))) == NULL)
+		goto failed;
 	TAILQ_INIT(ifs);
 
 	for (ifa = ifaddrs; ifa; ifa = ifa->ifa_next) {
@@ -466,7 +460,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 			}
 			ifp->hwlen = sdl->sdl_alen;
 #ifndef CLLADDR
-#  define CLLADDR(s) ((const char *)((s)->sdl_data + (s)->sdl_nlen))
+#  define CLLADDR(s) (const void *)((s)->sdl_data + (s)->sdl_nlen)
 #endif
 			memcpy(ifp->hwaddr, CLLADDR(sdl), ifp->hwlen);
 #elif AF_PACKET
@@ -554,16 +548,9 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 		TAILQ_INSERT_TAIL(ifs, ifp, next);
 	}
 
-#ifdef GETIFADDRS_AFLINK
-	{
-#else
+	if_learnaddrs(ctx, ifs, ifaddrs);
+failed:
 	freeifaddrs(ifaddrs);
-	if (getifaddrs(&ifaddrs) != -1) {
-#endif
-		if_learnaddrs(ctx, ifs, ifaddrs);
-		freeifaddrs(ifaddrs);
-	}
-
 	return ifs;
 }
 
