@@ -113,7 +113,7 @@ ipv4_getnetmask(uint32_t addr)
 
 struct ipv4_addr *
 ipv4_iffindaddr(struct interface *ifp,
-    const struct in_addr *addr, const struct in_addr *net)
+    const struct in_addr *addr, const struct in_addr *mask)
 {
 	struct ipv4_state *state;
 	struct ipv4_addr *ap;
@@ -122,7 +122,7 @@ ipv4_iffindaddr(struct interface *ifp,
 	if (state) {
 		TAILQ_FOREACH(ap, &state->addrs, next) {
 			if ((addr == NULL || ap->addr.s_addr == addr->s_addr) &&
-			    (net == NULL || ap->net.s_addr == net->s_addr))
+			    (mask == NULL || ap->mask.s_addr == mask->s_addr))
 				return ap;
 		}
 	}
@@ -154,8 +154,8 @@ ipv4_iffindmaskaddr(struct interface *ifp, const struct in_addr *addr)
 	state = IPV4_STATE(ifp);
 	if (state) {
 		TAILQ_FOREACH (ap, &state->addrs, next) {
-			if ((ap->addr.s_addr & ap->net.s_addr) ==
-			    (addr->s_addr & ap->net.s_addr))
+			if ((ap->addr.s_addr & ap->mask.s_addr) ==
+			    (addr->s_addr & ap->mask.s_addr))
 				return ap;
 		}
 	}
@@ -203,31 +203,33 @@ ipv4_srcaddr(const struct rt *rt, struct in_addr *addr)
 
 	/* Prefer DHCP source address if matching */
 	dstate = D_CSTATE(rt->iface);
-	if (dstate &&
-	    rt->net.s_addr == dstate->net.s_addr &&
-	    rt->dest.s_addr == (dstate->addr.s_addr & dstate->net.s_addr))
+	if (dstate && dstate->addr &&
+	    rt->mask.s_addr == dstate->addr->mask.s_addr &&
+	    rt->dest.s_addr ==
+	    (dstate->addr->addr.s_addr & dstate->addr->mask.s_addr))
 	{
-		*addr = dstate->addr;
+		*addr = dstate->addr->addr;
 		return 1;
 	}
 
 	/* Then IPv4LL source address if matching */
 	istate = IPV4LL_CSTATE(rt->iface);
-	if (istate &&
-	    rt->net.s_addr == inaddr_llmask.s_addr &&
-	    rt->dest.s_addr == (istate->addr.s_addr & inaddr_llmask.s_addr))
+	if (istate && istate->addr &&
+	    rt->mask.s_addr == istate->addr->mask.s_addr &&
+	    rt->dest.s_addr ==
+	    (istate->addr->addr.s_addr & istate->addr->mask.s_addr))
 	{
-		*addr = istate->addr;
+		*addr = istate->addr->addr;
 		return 1;
 	}
 
 	/* If neither match, return DHCP then IPv4LL */
-	if (dstate) {
-		*addr = dstate->addr;
+	if (dstate && dstate->addr) {
+		*addr = dstate->addr->addr;
 		return 0;
 	}
-	if (istate) {
-		*addr = istate->addr;
+	if (istate && istate->addr) {
+		*addr = istate->addr->addr;
 		return 0;
 	}
 
@@ -245,8 +247,8 @@ ipv4_hasaddr(const struct interface *ifp)
 	istate = IPV4LL_CSTATE(ifp);
 	return ((dstate &&
 	    dstate->added == STATE_ADDED &&
-	    dstate->addr.s_addr != INADDR_ANY) ||
-	    (istate && istate->addr.s_addr != INADDR_ANY));
+	    dstate->addr != NULL) ||
+	    (istate && istate->addr));
 }
 
 void
@@ -278,25 +280,6 @@ ipv4_init(struct dhcpcd_ctx *ctx)
 	return 0;
 }
 
-int
-ipv4_protocol_fd(const struct interface *ifp, uint16_t protocol)
-{
-
-	if (protocol == ETHERTYPE_ARP) {
-		const struct iarp_state *istate;
-
-		istate = ARP_CSTATE(ifp);
-		assert(istate != NULL);
-		return istate->fd;
-	} else {
-		const struct dhcp_state *dstate;
-
-		dstate = D_CSTATE(ifp);
-		assert(dstate != NULL);
-		return dstate->raw_fd;
-	}
-}
-
 /* Interface comparer for working out ordering. */
 int
 ipv4_ifcmp(const struct interface *si, const struct interface *ti)
@@ -346,7 +329,7 @@ find_route(struct rt_head *rts, const struct rt *r, const struct rt *srt)
 		    rt->iface->metric == r->iface->metric)) &&
 #endif
                     (!srt || srt != rt) &&
-		    rt->net.s_addr == r->net.s_addr)
+		    rt->mask.s_addr == r->mask.s_addr)
 			return rt;
 	}
 	return NULL;
@@ -360,28 +343,28 @@ desc_route(const char *cmd, const struct rt *rt)
 	const char *ifname = rt->iface ? rt->iface->name : NULL;
 
 	strlcpy(addr, inet_ntoa(rt->dest), sizeof(addr));
-	if (rt->net.s_addr == htonl(INADDR_BROADCAST) &&
+	if (rt->mask.s_addr == htonl(INADDR_BROADCAST) &&
 	    rt->gate.s_addr == htonl(INADDR_ANY))
 		logger(ctx, LOG_INFO, "%s: %s host route to %s",
 		    ifname, cmd, addr);
-	else if (rt->net.s_addr == htonl(INADDR_BROADCAST))
+	else if (rt->mask.s_addr == htonl(INADDR_BROADCAST))
 		logger(ctx, LOG_INFO, "%s: %s host route to %s via %s",
 		    ifname, cmd, addr, inet_ntoa(rt->gate));
 	else if (rt->dest.s_addr == htonl(INADDR_ANY) &&
-	    rt->net.s_addr == htonl(INADDR_ANY) &&
+	    rt->mask.s_addr == htonl(INADDR_ANY) &&
 	    rt->gate.s_addr == htonl(INADDR_ANY))
 		logger(ctx, LOG_INFO, "%s: %s default route",
 		    ifname, cmd);
 	else if (rt->gate.s_addr == htonl(INADDR_ANY))
 		logger(ctx, LOG_INFO, "%s: %s route to %s/%d",
-		    ifname, cmd, addr, inet_ntocidr(rt->net));
+		    ifname, cmd, addr, inet_ntocidr(rt->mask));
 	else if (rt->dest.s_addr == htonl(INADDR_ANY) &&
-	    rt->net.s_addr == htonl(INADDR_ANY))
+	    rt->mask.s_addr == htonl(INADDR_ANY))
 		logger(ctx, LOG_INFO, "%s: %s default route via %s",
 		    ifname, cmd, inet_ntoa(rt->gate));
 	else
 		logger(ctx, LOG_INFO, "%s: %s route to %s/%d via %s",
-		    ifname, cmd, addr, inet_ntocidr(rt->net),
+		    ifname, cmd, addr, inet_ntocidr(rt->mask),
 		    inet_ntoa(rt->gate));
 }
 
@@ -400,7 +383,7 @@ ipv4_findrt(struct dhcpcd_ctx *ctx, const struct rt *rt, int flags)
 #else
 		    (!flags || rt->iface == r->iface) &&
 #endif
-		    rt->net.s_addr == r->net.s_addr)
+		    rt->mask.s_addr == r->mask.s_addr)
 			return r;
 	}
 	return NULL;
@@ -464,7 +447,7 @@ nc_route(struct rt *ort, struct rt *nrt)
 
 	/* Don't set default routes if not asked to */
 	if (nrt->dest.s_addr == 0 &&
-	    nrt->net.s_addr == 0 &&
+	    nrt->mask.s_addr == 0 &&
 	    !(nrt->iface->options->options & DHCPCD_GATEWAY))
 		return -1;
 
@@ -491,7 +474,7 @@ nc_route(struct rt *ort, struct rt *nrt)
 	    ort->metric == nrt->metric &&
 #endif
 	    ort->dest.s_addr == nrt->dest.s_addr &&
-	    ort->net.s_addr ==  nrt->net.s_addr &&
+	    ort->mask.s_addr ==  nrt->mask.s_addr &&
 	    ort->gate.s_addr == nrt->gate.s_addr)
 	{
 		if (ort->mtu == nrt->mtu)
@@ -559,19 +542,23 @@ d_route(struct rt *rt)
 static struct rt_head *
 add_subnet_route(struct rt_head *rt, const struct interface *ifp)
 {
-	const struct dhcp_state *s;
+	const struct dhcp_state *state;
 	struct rt *r;
 
 	if (rt == NULL) /* earlier malloc failed */
 		return NULL;
 
-	s = D_CSTATE(ifp);
+	/* P2P interfaces don't have subnet routes as such. */
+	if (ifp->flags & IFF_POINTOPOINT)
+		return rt;
+
+	state = D_CSTATE(ifp);
 	/* Don't create a subnet route for these addresses */
-	if (s->net.s_addr == INADDR_ANY)
+	if (state->addr->mask.s_addr == INADDR_ANY)
 		return rt;
 #ifndef BSD
 	/* BSD adds a route in this instance */
-	if (s->net.s_addr == INADDR_BROADCAST)
+	if (state->addr->mask.s_addr == INADDR_BROADCAST)
 		return rt;
 #endif
 
@@ -580,11 +567,11 @@ add_subnet_route(struct rt_head *rt, const struct interface *ifp)
 		ipv4_freeroutes(rt);
 		return NULL;
 	}
-	r->dest.s_addr = s->addr.s_addr & s->net.s_addr;
-	r->net.s_addr = s->net.s_addr;
+	r->dest.s_addr = state->addr->addr.s_addr & state->addr->mask.s_addr;
+	r->mask.s_addr = state->addr->mask.s_addr;
 	r->gate.s_addr = INADDR_ANY;
 	r->mtu = dhcp_get_mtu(ifp);
-	r->src = s->addr;
+	r->src = state->addr->addr;
 
 	TAILQ_INSERT_HEAD(rt, r, next);
 	return rt;
@@ -595,13 +582,13 @@ static struct rt_head *
 add_loopback_route(struct rt_head *rt, const struct interface *ifp)
 {
 	struct rt *r;
-	const struct dhcp_state *s;
+	const struct dhcp_state *state;
 
 	if (rt == NULL) /* earlier malloc failed */
 		return NULL;
 
-	s = D_CSTATE(ifp);
-	if (s->addr.s_addr == INADDR_ANY)
+	state = D_CSTATE(ifp);
+	if (state->addr == NULL)
 		return rt;
 
 	if ((r = calloc(1, sizeof(*r))) == NULL) {
@@ -609,11 +596,11 @@ add_loopback_route(struct rt_head *rt, const struct interface *ifp)
 		ipv4_freeroutes(rt);
 		return NULL;
 	}
-	r->dest = s->addr;
-	r->net.s_addr = INADDR_BROADCAST;
+	r->dest = state->addr->addr;
+	r->mask.s_addr = INADDR_BROADCAST;
 	r->gate.s_addr = htonl(INADDR_LOOPBACK);
 	r->mtu = dhcp_get_mtu(ifp);
-	r->src = s->addr;
+	r->src = state->addr->addr;
 	TAILQ_INSERT_HEAD(rt, r, next);
 	return rt;
 }
@@ -648,7 +635,7 @@ get_routes(struct interface *ifp)
 	if (nrt) {
 		state = D_CSTATE(ifp);
 		TAILQ_FOREACH(rt, nrt, next) {
-			rt->src = state->addr;
+			rt->src = state->addr->addr;
 		}
 	}
 
@@ -673,10 +660,10 @@ add_destination_route(struct rt_head *rt, const struct interface *ifp)
 		return NULL;
 	}
 	r->dest.s_addr = INADDR_ANY;
-	r->net.s_addr = INADDR_ANY;
-	r->gate.s_addr = state->brd.s_addr;
+	r->mask.s_addr = INADDR_ANY;
+	r->gate = state->addr->brd;
 	r->mtu = dhcp_get_mtu(ifp);
-	r->src = state->addr;
+	r->src = state->addr->addr;
 	TAILQ_INSERT_HEAD(rt, r, next);
 	return rt;
 }
@@ -694,6 +681,10 @@ add_router_host_route(struct rt_head *rt, const struct interface *ifp)
 	if (rt == NULL) /* earlier malloc failed */
 		return NULL;
 
+	/* Don't add a host route for these interfaces. */
+	if (ifp->flags & (IFF_LOOPBACK | IFF_POINTOPOINT))
+		return rt;
+
 	TAILQ_FOREACH(rtp, rt, next) {
 		if (rtp->dest.s_addr != INADDR_ANY)
 			continue;
@@ -707,8 +698,8 @@ add_router_host_route(struct rt_head *rt, const struct interface *ifp)
 			/* match subnet */
 			cp = (const char *)&rtp->gate.s_addr;
 			cp2 = (const char *)&rtn->dest.s_addr;
-			cp3 = (const char *)&rtn->net.s_addr;
-			cplim = cp3 + sizeof(rtn->net.s_addr);
+			cp3 = (const char *)&rtn->mask.s_addr;
+			cplim = cp3 + sizeof(rtn->mask.s_addr);
 			while (cp3 < cplim) {
 				if ((*cp++ ^ *cp2++) & *cp3++)
 					break;
@@ -747,10 +738,10 @@ add_router_host_route(struct rt_head *rt, const struct interface *ifp)
 			return NULL;
 		}
 		rtn->dest.s_addr = rtp->gate.s_addr;
-		rtn->net.s_addr = htonl(INADDR_BROADCAST);
+		rtn->mask.s_addr = htonl(INADDR_BROADCAST);
 		rtn->gate.s_addr = htonl(INADDR_ANY);
 		rtn->mtu = dhcp_get_mtu(ifp);
-		rtn->src = state->addr;
+		rtn->src = state->addr->addr;
 		TAILQ_INSERT_BEFORE(rtp, rtn, next);
 	}
 	return rt;
@@ -903,49 +894,42 @@ ipv4_buildroutes(struct dhcpcd_ctx *ctx)
 }
 
 int
-ipv4_deladdr(struct interface *ifp,
-    const struct in_addr *addr, const struct in_addr *net, int keeparp)
+ipv4_deladdr(struct ipv4_addr *addr, int keeparp)
 {
-	struct dhcp_state *dstate;
 	int r;
 	struct ipv4_state *state;
 	struct ipv4_addr *ap;
 	struct arp_state *astate;
-	uint32_t a, n;
 
-	logger(ifp->ctx, LOG_DEBUG,
-	    "%s: deleting IP address %s/%d",
-	    ifp->name, inet_ntoa(*addr), inet_ntocidr(*net));
+	logger(addr->iface->ctx, LOG_DEBUG,
+	    "%s: deleting IP address %s", addr->iface->name, addr->saddr);
 
-	r = if_deladdress(ifp, addr, net);
+	r = if_address(RTM_DELADDR, addr);
 	if (r == -1 && errno != EADDRNOTAVAIL && errno != ENXIO &&
 	    errno != ENODEV)
-		logger(ifp->ctx, LOG_ERR, "%s: %s: %m", ifp->name, __func__);
+		logger(addr->iface->ctx, LOG_ERR, "%s: %s: %m",
+		    addr->iface->name, __func__);
 
-	if (!keeparp && (astate = arp_find(ifp, addr)) != NULL)
+	if (!keeparp && (astate = arp_find(addr->iface, &addr->addr)) != NULL)
 		arp_free(astate);
 
-	a = addr->s_addr;
-	n = net->s_addr;
-	state = IPV4_STATE(ifp);
+	state = IPV4_STATE(addr->iface);
 	TAILQ_FOREACH(ap, &state->addrs, next) {
-		if (ap->addr.s_addr == addr->s_addr &&
-		    ap->net.s_addr == net->s_addr)
-		{
+		if (IPV4_MASK_EQ(ap, addr)) {
+			struct dhcp_state *dstate;
+
+			dstate = D_STATE(ap->iface);
 			TAILQ_REMOVE(&state->addrs, ap, next);
 			free(ap);
+
+			if (dstate && dstate->addr == ap) {
+				dstate->added = 0;
+				dstate->addr = NULL;
+			}
 			break;
 		}
 	}
 
-	/* Have to do this last incase the function arguments
-	 * were these very pointers. */
-	dstate = D_STATE(ifp);
-	if (dstate && dstate->addr.s_addr == a && dstate->net.s_addr == n) {
-		dstate->added = 0;
-		dstate->addr.s_addr = 0;
-		dstate->net.s_addr = 0;
-	}
 	return r;
 }
 
@@ -961,7 +945,7 @@ delete_address(struct interface *ifp)
 	if (ifo->options & DHCPCD_INFORM ||
 	    (ifo->options & DHCPCD_STATIC && ifo->req_addr.s_addr == 0))
 		return 0;
-	r = ipv4_deladdr(ifp, &state->addr, &state->net, 0);
+	r = ipv4_deladdr(state->addr, 0);
 	return r;
 }
 
@@ -994,7 +978,6 @@ ipv4_addaddr(struct interface *ifp, const struct in_addr *addr,
 {
 	struct ipv4_state *state;
 	struct ipv4_addr *ia;
-	char bcast_str[INET_ADDRSTRLEN];
 
 	if ((state = ipv4_getstate(ifp)) == NULL) {
 		logger(ifp->ctx, LOG_ERR, "%s: ipv4_getstate: %m", __func__);
@@ -1005,7 +988,7 @@ ipv4_addaddr(struct interface *ifp, const struct in_addr *addr,
 
 		TAILQ_FOREACH_SAFE(ia, &state->addrs, next, ian) {
 			if (ia->addr.s_addr != addr->s_addr)
-				ipv4_deladdr(ifp, &ia->addr, &ia->net, 0);
+				ipv4_deladdr(ia, 0);
 		}
 	}
 
@@ -1014,10 +997,18 @@ ipv4_addaddr(struct interface *ifp, const struct in_addr *addr,
 		return NULL;
 	}
 
-	strlcpy(bcast_str, inet_ntoa(*bcast), sizeof(bcast_str));
-	logger(ifp->ctx, LOG_DEBUG, "%s: adding IP address %s/%d broadcast %s",
-	    ifp->name, inet_ntoa(*addr), inet_ntocidr(*mask), bcast_str);
-	if (if_addaddress(ifp, addr, mask, bcast) == -1) {
+	ia->iface = ifp;
+	ia->addr = *addr;
+	ia->mask = *mask;
+	ia->brd = *bcast;
+#ifdef IN_IFF_TENTATIVE
+	ia->addr_flags = IN_IFF_TENTATIVE;
+#endif
+	snprintf(ia->saddr, sizeof(ia->saddr), "%s/%d",
+	    inet_ntoa(*addr), inet_ntocidr(*mask));
+	logger(ifp->ctx, LOG_DEBUG, "%s: adding IP address %s broadcast %s",
+	    ifp->name, ia->saddr, inet_ntoa(*bcast));
+	if (if_address(RTM_NEWADDR, ia) == -1) {
 		if (errno != EEXIST)
 			logger(ifp->ctx, LOG_ERR, "%s: if_addaddress: %m",
 			    __func__);
@@ -1025,13 +1016,6 @@ ipv4_addaddr(struct interface *ifp, const struct in_addr *addr,
 		return NULL;
 	}
 
-	ia->iface = ifp;
-	ia->addr = *addr;
-	ia->net = *mask;
-	ia->brd = *bcast;
-#ifdef IN_IFF_TENTATIVE
-	ia->addr_flags = IN_IFF_TENTATIVE;
-#endif
 	TAILQ_INSERT_TAIL(&state->addrs, ia, next);
 	return ia;
 }
@@ -1040,16 +1024,15 @@ static int
 ipv4_daddaddr(struct interface *ifp, const struct dhcp_lease *lease)
 {
 	struct dhcp_state *state;
+	struct ipv4_addr *ia;
 
-	if (ipv4_addaddr(ifp, &lease->addr, &lease->net, &lease->brd) == NULL)
+	ia = ipv4_addaddr(ifp, &lease->addr, &lease->mask, &lease->brd);
+	if (ia == NULL)
 		return -1;
 
 	state = D_STATE(ifp);
 	state->added = STATE_ADDED;
-
-	state->addr.s_addr = lease->addr.s_addr;
-	state->net.s_addr = lease->net.s_addr;
-
+	state->addr = ia;
 	return 0;
 }
 
@@ -1072,7 +1055,8 @@ ipv4_preferanother(struct interface *ifp)
 			break; /* We are already the most preferred */
 		nstate = D_STATE(ifn);
 		if (nstate && !nstate->added &&
-		    nstate->lease.addr.s_addr == state->addr.s_addr)
+		    state->addr != NULL &&
+		    nstate->lease.addr.s_addr == state->addr->addr.s_addr)
 		{
 			preferred = 1;
 			delete_address(ifp);
@@ -1098,7 +1082,7 @@ ipv4_applyaddr(void *arg)
 	struct dhcp_state *state = D_STATE(ifp), *nstate;
 	struct dhcp_lease *lease;
 	struct if_options *ifo = ifp->options;
-	struct ipv4_addr *ap;
+	struct ipv4_addr *ia;
 	int r;
 
 	if (state == NULL)
@@ -1130,7 +1114,8 @@ ipv4_applyaddr(void *arg)
 		}
 		nstate = D_STATE(ifn);
 		if (nstate && nstate->added &&
-		    nstate->addr.s_addr == lease->addr.s_addr)
+		    nstate->addr &&
+		    nstate->addr->addr.s_addr == lease->addr.s_addr)
 		{
 			if (r == 0) {
 				logger(ifp->ctx, LOG_INFO,
@@ -1144,7 +1129,7 @@ ipv4_applyaddr(void *arg)
 			    ifn->name,
 			    inet_ntoa(lease->addr),
 			    ifp->name);
-			ipv4_deladdr(ifn, &nstate->addr, &nstate->net, 0);
+			ipv4_deladdr(nstate->addr, 0);
 			break;
 		}
 	}
@@ -1154,64 +1139,63 @@ ipv4_applyaddr(void *arg)
 		TAILQ_FOREACH(ifn, ifp->ctx->ifaces, next) {
 			if (ifn == ifp)
 				continue;
-			ap = ipv4_iffindaddr(ifn, &lease->addr, NULL);
-			if (ap)
-				ipv4_deladdr(ifn, &ap->addr, &ap->net, 0);
+			ia = ipv4_iffindaddr(ifn, &lease->addr, NULL);
+			if (ia != NULL)
+				ipv4_deladdr(ia, 0);
 		}
 	}
 
 	/* If the netmask or broadcast is different, re-add the addresss */
-	ap = ipv4_iffindaddr(ifp, &lease->addr, NULL);
-	if (ap &&
-	    ap->net.s_addr == lease->net.s_addr &&
-	    ap->brd.s_addr == lease->brd.s_addr)
+	ia = ipv4_iffindaddr(ifp, &lease->addr, NULL);
+	if (ia &&
+	    ia->mask.s_addr == lease->mask.s_addr &&
+	    ia->brd.s_addr == lease->brd.s_addr)
 		logger(ifp->ctx, LOG_DEBUG,
-		    "%s: IP address %s/%d already exists",
-		    ifp->name, inet_ntoa(lease->addr),
-		    inet_ntocidr(lease->net));
+		    "%s: IP address %s already exists",
+		    ifp->name, ia->saddr);
 	else {
 #if __linux__
 		/* Linux does not change netmask/broadcast address
 		 * for re-added addresses, so we need to delete the old one
 		 * first. */
-		if (ap != NULL)
-			ipv4_deladdr(ifp, &ap->addr, &ap->net, 0);
+		if (ia != NULL)
+			ipv4_deladdr(ia, 0);
 #endif
 		r = ipv4_daddaddr(ifp, lease);
 		if (r == -1 && errno != EEXIST)
 			return;
 	}
 
-#ifdef IN_IFF_NOTUSEABLE
-	ap = ipv4_iffindaddr(ifp, &lease->addr, NULL);
-	if (ap == NULL) {
+	ia = ipv4_iffindaddr(ifp, &lease->addr, NULL);
+	if (ia == NULL) {
 		logger(ifp->ctx, LOG_ERR, "%s: added address vanished",
 		    ifp->name);
 		return;
-	} else if (ap->addr_flags & IN_IFF_NOTUSEABLE)
+	}
+#ifdef IN_IFF_NOTUSEABLE
+	if (ia->addr_flags & IN_IFF_NOTUSEABLE)
 		return;
 #endif
 
 	/* Delete the old address if different */
-	if (state->addr.s_addr != lease->addr.s_addr &&
-	    state->addr.s_addr != 0 &&
+	if (state->addr &&
+	    state->addr->addr.s_addr != lease->addr.s_addr &&
 	    ipv4_iffindaddr(ifp, &lease->addr, NULL))
 		delete_address(ifp);
 
+	state->addr = ia;
 	state->added = STATE_ADDED;
-	state->addr.s_addr = lease->addr.s_addr;
-	state->net.s_addr = lease->net.s_addr;
 
 	/* Find any freshly added routes, such as the subnet route.
 	 * We do this because we cannot rely on recieving the kernel
 	 * notification right now via our link socket. */
-	if_initrt(ifp);
+	if_initrt(ifp->ctx);
 	ipv4_buildroutes(ifp->ctx);
 	/* Announce the address */
 	if (ifo->options & DHCPCD_ARP) {
 		struct arp_state *astate;
 
-		if ((astate = arp_new(ifp, &state->addr)) != NULL)
+		if ((astate = arp_new(ifp, &state->addr->addr)) != NULL)
 			arp_announce(astate);
 	}
 	if (state->state == DHS_BOUND) {
@@ -1223,12 +1207,12 @@ ipv4_applyaddr(void *arg)
 void
 ipv4_handleifa(struct dhcpcd_ctx *ctx,
     int cmd, struct if_head *ifs, const char *ifname,
-    const struct in_addr *addr, const struct in_addr *net,
+    const struct in_addr *addr, const struct in_addr *mask,
     const struct in_addr *brd, int flags)
 {
 	struct interface *ifp;
 	struct ipv4_state *state;
-	struct ipv4_addr *ap;
+	struct ipv4_addr *ia;
 
 	if (ifs == NULL)
 		ifs = ctx->ifaces;
@@ -1236,10 +1220,6 @@ ipv4_handleifa(struct dhcpcd_ctx *ctx,
 		errno = ESRCH;
 		return;
 	}
-	if (addr->s_addr == INADDR_ANY) {
-		errno = EINVAL;
-		return;
-	}
 	if ((ifp = if_find(ifs, ifname)) == NULL)
 		return;
 	if ((state = ipv4_getstate(ifp)) == NULL) {
@@ -1247,43 +1227,58 @@ ipv4_handleifa(struct dhcpcd_ctx *ctx,
 		return;
 	}
 
-	ap = ipv4_iffindaddr(ifp, addr, NULL);
-	if (cmd == RTM_NEWADDR) {
-		if (ap == NULL) {
-			if ((ap = malloc(sizeof(*ap))) == NULL) {
+	ia = ipv4_iffindaddr(ifp, addr, NULL);
+	switch (cmd) {
+	case RTM_NEWADDR:
+		if (ia == NULL) {
+			if ((ia = malloc(sizeof(*ia))) == NULL) {
 				logger(ifp->ctx, LOG_ERR, "%s: %m", __func__);
 				return;
 			}
-			ap->iface = ifp;
-			ap->addr = *addr;
-			TAILQ_INSERT_TAIL(&state->addrs, ap, next);
-		}
-		ap->net = *net;
-		ap->brd = *brd;
-		ap->addr_flags = flags;
-	} else if (cmd == RTM_DELADDR) {
-		if (ap) {
-			TAILQ_REMOVE(&state->addrs, ap, next);
-			free(ap);
+			ia->iface = ifp;
+			ia->addr = *addr;
+			TAILQ_INSERT_TAIL(&state->addrs, ia, next);
 		}
+		/* Mask could have changed */
+		ia->mask = *mask;
+		snprintf(ia->saddr, sizeof(ia->saddr), "%s/%d",
+		    inet_ntoa(*addr), inet_ntocidr(*mask));
+		if (brd != NULL)
+			ia->brd = *brd;
+		else
+			ia->brd.s_addr = INADDR_ANY;
+		ia->addr_flags = flags;
+		break;
+	case RTM_DELADDR:
+		if (ia == NULL)
+			return;
+		TAILQ_REMOVE(&state->addrs, ia, next);
+		break;
+	default:
+		return;
 	}
 
-	arp_handleifa(cmd, ifp, addr, flags);
-	dhcp_handleifa(cmd, ifp, addr, net, brd, flags);
+	if (addr->s_addr != INADDR_ANY && addr->s_addr != INADDR_BROADCAST) {
+		arp_handleifa(cmd, ia);
+		dhcp_handleifa(cmd, ia);
+	}
+
+	if (cmd == RTM_DELADDR)
+		free(ia);
 }
 
 void
 ipv4_free(struct interface *ifp)
 {
 	struct ipv4_state *state;
-	struct ipv4_addr *addr;
+	struct ipv4_addr *ia;
 
 	if (ifp) {
 		state = IPV4_STATE(ifp);
 		if (state) {
-		        while ((addr = TAILQ_FIRST(&state->addrs))) {
-				TAILQ_REMOVE(&state->addrs, addr, next);
-				free(addr);
+		        while ((ia = TAILQ_FIRST(&state->addrs))) {
+				TAILQ_REMOVE(&state->addrs, ia, next);
+				free(ia);
 			}
 			ipv4_freerts(&state->routes);
 #ifdef BSD
